/**
 * The abstract joogl camera object.
 * <p>
 * This object is used to create a camera that can be put in the scene graph.
 * There are subtypes for the different camera types.  
 *
 * @version 	1.00
 * @author      Blair MacIntyre
 */
 
public abstract class jooglCamera extends jooglNode implements Cloneable {        
    /** the x coordinate of the eye (from) location */
    protected double fx;
    /** the y coordinate of the eye (from) location */
    protected double fy;
    /** the z coordinate of the eye (from) location */
    protected double fz;
    
    /** the x coordinate of the to location */
    protected double tx;
    /** the y coordinate of the to location */
    protected double ty;
    /** the z coordinate of the to location */
    protected double tz;
    
    /** the x coordinate of the up vector */
    protected double ux;
    /** the y coordinate of the up vector */
    protected double uy;
    /** the z coordinate of the up vector */
    protected double uz;

    /** the near clipping plane of the camera */
    protected double near;
    /** the far clipping plane of the camera */
    protected double far;

    /**
     * The aspect ratio of the camera view.
     * <p>
     * This is set automatically to be the aspect of the viewport.
     */
    double aspect;
    
    /** 
     * A constructor for the abstract camera.
     * <p>
     * Create a generic camera object to be used to view a scene graph.
     *
     * @param _fx  the x coordinate of the eyepoint
     * @param _fy  the y coordinate of the eyepoint
     * @param _fz  the z coordinate of the eyepoint
     * @param _tx  the x coordinate of the to point
     * @param _ty  the y coordinate of the to point
     * @param _tz  the z coordinate of the to point
     * @param _ux  the x coordinate of the up vector
     * @param _uy  the y coordinate of the up vector
     * @param _uz  the z coordinate of the up vector
     * @param _near  the near clipping plane
     * @param _far   the far clipping plane
     */
    public jooglCamera(double _fx, double _fy, double _fz, 
                        double _tx, double _ty, double _tz,
			double _ux, double _uy, double _uz,
			double _near, double _far)
    {
        super();
        fx = _fx;
        fy = _fy;
        fz = _fz;
        tx = _tx;
        ty = _ty;
        tz = _tz;
        ux = _ux;
        uy = _uy;
        uz = _uz;
        near = _near;
        far = _far;
        aspect = 1.0;
    }

    /** 
     * A constructor for the abstract camera using default parameters.
     * <p>
     * Create an abstract camera object to be used to view a scene graph.
     * Sets of the camera with an eyepoint of (0,0,0), a to point of (0,0,1),
     * an up vector of (0,1,0), a near clipping
     * plane of 1, and a far clipping plane of 1000.
     */
    public jooglCamera()
    {
        this(0.0, 0.0, 0.0, 
             0.0, 0.0, 1.0,
             0.0, 1.0, 0.0,  
             1.0, 1000.0);
    }

    /** 
     * Change the from (eye) point of the camera.
     *
     * @param _fx  the x coordinate of the eyepoint
     * @param _fy  the y coordinate of the eyepoint
     * @param _fz  the z coordinate of the eyepoint
     */
    public void setFrom(double _fx, double _fy, double _fz)
    {
        fx = _fx;
        fy = _fy;
        fz = _fz;
        redrawWindows();
    }
    
    /** 
     * Get the from (eye) point of the camera.
     *
     * @return a vector of 3 doubles containing the eyepoint
     */
    public double[] getFrom()
    {
        double[] f = new double[3];
        f[0] = fx;
        f[1] = fy;
        f[2] = fz;
        return f;
    }

    /** 
     * Change the to point of the camera.
     *
     * @param _tx  the x coordinate of the to point
     * @param _ty  the y coordinate of the to point
     * @param _tz  the z coordinate of the to point
     */
    public void setTo(double _tx, double _ty, double _tz)
    {
        tx = _tx;
        ty = _ty;
        tz = _tz;
        redrawWindows();
    }
    
    /** 
     * Get the to point of the camera.
     *
     * @return a vector of 3 doubles containing the to point
     */
    public double[] getTo()
    {
        double[] t = new double[3];
        t[0] = tx;
        t[1] = ty;
        t[2] = tz;
        return t;
    }

    /** 
     * Change the up vector of the camera.
     *
     * @param _ux  the x coordinate of the up vector
     * @param _uy  the y coordinate of the up vector
     * @param _uz  the z coordinate of the up vector
     */
    public void setUp(double _ux, double _uy, double _uz)
    {
        ux = _ux;
        uy = _uy;
        uz = _uz;
        redrawWindows();
    }
    
    /** 
     * Get the up vector of the camera.
     *
     * @return a vector of 3 doubles containing the up vector
     */
    public double[] getUp()
    {
        double[] u = new double[3];
        u[0] = ux;
        u[1] = uy;
        u[2] = uz;
        return u;
    }

    /** 
     * Set the near clipping plane of the camera.
     *
     * @param _near a double containing the near value
     */
    public void setNear(double _near)
    {
        near = _near;
        redrawWindows();
    }
    
    /** 
     * Get the near clipping plane of the camera.
     *
     * @return a double containing the near value
     */
    public double getNear()
    {
        return near;
    }
    
    /** 
     * Set the far clipping plane of the camera.
     *
     * @param _far a double containing the far value
     */
    public void setFar(double _far)
    {
        far = _far;
        redrawWindows();
    }

    /** 
     * Get the near clipping plane of the camera.
     *
     * @return a double containing the far value
     */
    public double getFar()
    {
        return far;
    }
    
    /**
     * Internal method, called to set the aspect ratio of the camera.
     *
     * @param a the aspect ration of the viewport.
     */
    protected void setAspect(double a)
    {
        aspect = a;
    }
    
    /**
     * Internal method, called to set the current PROJECTION to correspond to this camera.
     * <p>
     * The rendering engine calls this method with the transformation from the root to the
     * camera node in the scene graph.  The camera should be set up so that it is transformed
     * by this matrix to be in this location.
     *
     * @param camCoords  the transformation created by multiplying the transformations along the
     *                   path from the root to the camera node.
     */
    protected abstract void apply(jooglTransform camCoords);
}
