import java.awt.*;
import java.awt.event.*;
import gl4java.*;
import gl4java.utils.glut.*;
import gl4java.GLContext;
import gl4java.awt.GLAnimCanvas;

import java.util.Vector;
import java.util.Iterator;

/**
 * The joogl canvas object.   The window that is rendered into.
 * <p>
 * Create a canvas that has a scene graph rooted at it.  This is a subclass of
 * GLAnimCanvas, and uses the facilities of the AnimCanvas to render.
 * <p>
 * The canvas contains one or more jooglRoot nodes that are
 * the roots of scene graphs to be drawn in viewports on the canvas.
 * The graph attached to each root node is drawn in the canvas 
 * using the specified viewport, and the specified camera object for viewing.  
 * If the camera appears in the scene graph, it will be transformed
 * accordingly, but it does not need to be attached to the graph (in which case
 * it will use its parameters directly).
 * <p>
 * If there are any dynamic parameters in any of the root nodes attached
 * to the canvas, the AnimCanvas will draw continuously.  If there are no
 * dynamic properties, the animation is suspended.  
 * <p>
 * <b>NOTE:</b>  There appears to be a bug in GL4Java, that multiple canvases
 * do not work correctly.  For now, do not open multiple windows until you know
 * your software works.  It appears to be a problem with the locking in the paint
 * method, which allows the OpenGL commands from the multiple windows to be 
 * interleaved.
 *
 * @version 	1.00
 * @author      Blair MacIntyre
 */ 
public class jooglCanvas extends GLAnimCanvas 
{
    /** the roots of the scene graphs for this window */
    protected Vector roots = new Vector();
    
    /** the clock for this window, shared by all roots */
    protected jooglClock clock = new jooglRealTimeClock();
    
    /** the width of this window */
    protected int width;
    /** the height of this window */
    protected int height;

    /** an jooglTransform used to set default state when rendering */
    protected jooglTransform defaultTransform = new jooglTransform();
    /** an jooglMaterial used to set default state when rendering */
    protected jooglMaterial defaultMaterial = new jooglMaterial();
    /** an jooglDrawMode used to set default state when rendering */
    protected jooglDrawMode defaultDrawMode = new jooglDrawMode();
    /** an jooglShadeMode used to set default state when rendering */
    protected jooglShadeMode defaultShadeMode = new jooglShadeMode();
    
    // the visitors for drawing the scene graph
    // 
    /** Used to visit the graph, updating the camera and lights, and
      * update dynamic properties and damaged geometry nodes */
    protected jooglLightsCameraActionVisitor lightsCameraActionVisitor = new jooglLightsCameraActionVisitor();
    /** Used to visit the graph, rending the opaque geometry */
    protected jooglRenderVisitor renderVisitor = new jooglRenderVisitor();
    /** Used to visit the graph, rending the transparent geometry */
    protected jooglRenderTransparentVisitor renderTransparentVisitor = 
            new jooglRenderTransparentVisitor();
        
    /** Is this window dynamically changing, or static? If dynamic, redraw continuously */
    protected boolean dynamic = false;

    /**
     * A constructor that specifies the graphics configuration, capabilities, width and height.
     *
     * @param g The desired graphics configuration
     * @param glCaps The desired graphics capabilities
     * @param w The width of the window
     * @param h The height of the window
     */
    public jooglCanvas(GraphicsConfiguration g, GLCapabilities glCaps, int w, int h)
    {
        super(g, glCaps, w, h);
        height = h;
        width = w;
    }
    
    /**
     * A constructor that specifies the capabilities, width and height.
     *
     * @param glCaps The desired graphics capabilities
     * @param w The width of the window
     * @param h The height of the window
     */
    public jooglCanvas(GLCapabilities glCaps, int w, int h)
    {
        super(glCaps, w, h);
        height = h;
        width = w;
    }

    /**
     * A constructor that specifies the width and height.
     *
     * @param w The width of the window
     * @param h The height of the window
     */
    public jooglCanvas(int w, int h)
    {
        super(w, h);
        height = h;
        width = w;
    }

    /**
     * Set the clock for this window.
     * <p>
     * This is the clock that will be used by all jooglRoot objects to render into this window.
     *
     * @param c The new clock object
     */
    public void setClock(jooglClock c)
    {
        clock = c;
        repaint();
    }
    
    /**
     * Get the clock for this window.
     * <p>
     * This is the clock that is being used by all jooglRoot objects to render into this window.
     *
     * @return The current clock object
     */
    public jooglClock getClock()
    {
        return clock;
    }
    
    // methods to manipulate the collection of viewports.  
    // These are analogous to the Collection methods.
    
    /**
     * Add a jooglRoot to the canvas.
     *  <p>
     * Analogous to the add method on a Vector object.
     *
     * @param n the new root object
     * @return true if the collection of Roots changed as a result of the call
     */
    public boolean addRoot(jooglRoot n) 
    {
        boolean ret = roots.add(n);
        repaint();
        return ret;
    }
    
    /**
     * remove a jooglRoot from the canvas.
     *  <p>
     * Analogous to the remove method on a Collection object.
     *
     * @param n the root object to remove
     * @return true if the collection of Roots contained n
     */
    public boolean removeRoot(jooglRoot n) 
    {
        boolean ret = roots.remove(n);
        repaint();
        return ret;
    }
        
    /**
     * Remove all jooglRoot objects from the canvas.
     *  <p>
     * Analogous to the clear method on a Collection object.
     */
    public void clearRoots()
    {
        roots.clear();
        repaint();
    }
    
    /**
     * Test to see if a jooglRoot is in the canvas.
     *  <p>
     * Analogous to the contains method on a Collection object.
     *
     * @param n the root object
     * @return true if the collection of Roots contains n
     */
    public boolean containsRoot(jooglRoot n)
    {
        return roots.contains(n);
    }
        
    /**
     * Create an iterator over the collection of jooglRoots in the canvas.
     *  <p>
     * Analogous to the interator method on a Collection object.
     *
     * @return a new iterator on the collection of Root objects
     */
    public Iterator rootIterator() 
    {
        return roots.iterator();
    }


    /**
     * Test if there are any jooglRoots in the canvas.
     *  <p>
     * Analogous to the isEmpty method on a Collection object.
     *
     * @return true if there is at least one Roots in the canvas
     */
    public boolean isRoots()
    {
        return !roots.isEmpty();
    }
    
    /**
     * The number of roots in the canvas.
     *  <p>
     * Analogous to the size method on a Collection object.
     *
     * @return the number of Root objects in the canvas
     */
    public int numberOfRoots() 
    {
        return roots.size();
    }
        
    /**
    * Pick objects at position x,y in the root's window, with an
    * area of width/height around the mouse point.  
    * <p>
    * pickRoot() returns a Vector filled with the hit records for each object
    * hit.  If there are no objects hit, pickRoot() returns null. Each hit
    * record is a Vector, with the contents as follows: 
    * <pre>
       [jooglRoot, number of objects on the branch, zmin, zmax, o, p1, p2,...,pN] 
    * </pre>
    * o is the geometry object picked. The pi's are jooglPickable property nodes, 
    * and correspond to the pickable objects on the path
    * from the root node, starting with the object closest to the root (p1)
    * and ending with the object closest to geometry node (pN).   Only those pickable
    * properties with pickable set to true will be included.
    *
    * @param x  the x location of the selector (cursor)
    * @param y  the y location of the selector (cursor)
    * @param width  the width of the area around the cursor
    * @param height  the height of the area around the cursor
    */
    public Vector pickRoot (double x, double y, double width, double height)
    {
        return null;
    }
   
    /**
     * Initialize the canvas before creating the GLContext.
     * <p>
     * Called just BEFORE the GL-Context is created.
     * <p>
     * Define a reasonable set of 3D options of the jooglCanvas.  
     * Also, attempt to set the sharedGLContext here, so that all jooglCanvas
     * objects share textures and display lists.
     */
    public void preInit()
    {
        // We want double buffering
        doubleBuffer = true;
        // But we dont want stereo view
        stereoView = false; 
        
        // We want to share a glcontext, so display lists and textures are shared
        // we can call getSharedContext() on any joogl object
        sharedGLContext = clock.getSharedContext();
        //System.out.println("preInit: " + sharedGLContext);
    }

    /**
     * Initialize the canvas after creating the GLContext.
     * <p>
     * Called just AFTER the GL-Context is created.   Set up the OpenGL state for joogl.
     */
    public void init()
    {
        //System.out.println("init: " + sharedGLContext);
        //System.out.println("init: " + glj);

        setAnimateFps(35);
        
        width = getSize().width;
        height = getSize().height;
        
        // enable backface culling
        gl.glEnable(GL_CULL_FACE);
        gl.glDepthFunc(GL_LESS);
        gl.glEnable(GL_DEPTH_TEST);
        gl.glFrontFace(GL_CCW);
        gl.glEnable(GL_AUTO_NORMAL);
        gl.glEnable(GL_NORMALIZE);    
        gl.glEnable(GL_COLOR_MATERIAL);
        gl.glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
                        
        /* set up the default viewing stacks */
        gl.glMatrixMode(GL_PROJECTION);
        gl.glLoadIdentity();
        gl.glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);
        gl.glMatrixMode(GL_MODELVIEW);
        gl.glLoadIdentity();

        /* enable lighting */
        gl.glEnable(GL_LIGHTING);

        /* turn off all lights, just to be safe! */
        int[] maxLight = new int[1];
        gl.glGetIntegerv(GL_MAX_LIGHTS, maxLight);
        for (int i=maxLight[0]-1; i>=0; i--)
            gl.glDisable(GL_LIGHT0+i);

        /* turn off the global ambient light */
        gl.glLightModelfv(GL_LIGHT_MODEL_AMBIENT, jooglLight.zeroLight);

        /* enable local viewer calculations on the lights */
        gl.glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);

        // add this canvas to the list of known jooglCanvases
        clock.addCanvas(this);

        // if we did not set a sharedGLContext, our canvas should be the shared context
        if (sharedGLContext == null)
            clock.setSharedContext(getGLContext());
    }
        
    /**
     * Destroy the canvas.
     */
    public void cvsDispose()
    {
        clock.removeCanvas(this);
    }
	
    /**
     * Called after the first paint command, and any window resize. 
     * <p>
     * In addition to keeping track of the window size, tell any root's to resize their viewports if necessary. 
     *
     * @param w the width of the window
     * @param h the height of the window
     */
    public void reshape(int w, int h)
    {
        // Reset The Current Viewport And Perspective Transformation
        width = w;
        height = h;
        //gl.glViewport(0, 0, width, height);
        
        Iterator i = roots.iterator();
        while (i.hasNext()) 
        {
            ((jooglRoot)(i.next())).adjustToWindowReshape(width, height);

        }

        //aspect = ((double)w)/((double)h);
        //cam.setAspect(aspect);
    }
        
    /**
     * Render all the jooglRoots.
     * <p>
     * Draw to the canvas.   All of the work to draw the jooglRoot objects should be done here.
     * 
     */
    public void display()
    {
        // Ensure GL is initialised correctly
        if (glj.gljMakeCurrent() == false)
            return;
        
        // System.out.println("display: " + glj);

        // Clear The Whole Window And The Depth Buffer
        gl.glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
        gl.glViewport(0, 0, width, height);
        gl.glClear(GL_COLOR_BUFFER_BIT| GL_DEPTH_BUFFER_BIT);
        
        dynamic = false;
        long time = clock.getTime();
        
        Iterator i = roots.iterator();
        while (i.hasNext()) 
        {
            jooglRoot root = ((jooglRoot)(i.next()));
            root.setViewport();
            
            //System.out.println("display root: " + root);
            //System.out.println("gl.glClearColor("+root.bgRed+", "+root.bgGreen+", "+root.bgBlue+");");
            
            // render the background.
            // disable the depth testing and writing into the z-buffer
            gl.glDepthMask(GL_FALSE);
            gl.glDisable(GL_DEPTH_TEST);
            
            // turn off lighting so we use the exact color given
            gl.glDisable(GL_LIGHTING);
    
            // set a nice default projection
            gl.glMatrixMode(GL_PROJECTION);
            gl.glLoadIdentity();
            gl.glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);

            // draw a polygon of the correct color
            gl.glColor3f(root.bgRed, root.bgGreen, root.bgBlue);
            gl.glBegin(GL_POLYGON);
               gl.glVertex3f(-1.0f, -1.0f, 0.0f);
               gl.glVertex3f(1.0f, -1.0f, 0.0f);
               gl.glVertex3f(1.0f, 1.0f, 0.0f);
               gl.glVertex3f(-1.0f, 1.0f, 0.0f);
            gl.glEnd();
            // reenable the depth testing and lighting
            gl.glDepthMask(GL_TRUE);
            gl.glEnable(GL_DEPTH_TEST);
            gl.glEnable(GL_LIGHTING);

            // switch back to modelview matrix manipulation
            gl.glMatrixMode(GL_MODELVIEW);

            // push the default values onto the stacks, which also sets their values
            defaultTransform.push();
            defaultMaterial.push();
            defaultDrawMode.push();
            defaultShadeMode.push();
        
            // set the first light to be LIGHT0, mark the current camera, and set the time to use for 
            // property recomputation
            lightsCameraActionVisitor.setLightCameraTime(GL_LIGHT0, root.cam, time);
            // set the camera projection based on it not being in the scene graph
            root.cam.apply(new jooglTransform());
            // visit the graph, creating the lights and setting the camera projection
            // based on the location of the camera node (if it's in the graph).
            lightsCameraActionVisitor.visit(root);
   
            // if any of the viewports are dynamic, the whole window is
            dynamic = dynamic || (clock.isMoving() && lightsCameraActionVisitor.isDynamic());

            // visit the scene graph, rendering the nodes as we hit them.
            //System.out.println(" render geom");
            renderVisitor.visit(root);
            gl.glEnable(GL_BLEND);
            gl.glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
            gl.glDepthMask(GL_FALSE);
            //System.out.println(" render transparent geom");
            renderTransparentVisitor.visit(root);
            gl.glDepthMask(GL_TRUE);
            gl.glDisable(GL_BLEND);
        
            // pop the default values off the stacks
            defaultTransform.pop();
            defaultMaterial.pop();
            defaultDrawMode.pop();
            defaultShadeMode.pop();
        
            // turn off all the lights, so we are ready for any changes that might happen next frame
            if (lightsCameraActionVisitor.getNextLight() > GL_LIGHT0)
                for (int light = GL_LIGHT0; light < lightsCameraActionVisitor.getNextLight(); light++)
                    gl.glDisable(light);
            
    /*
            // render the background.
            // disable the depth testing and writing into the z-buffer
            gl.glDepthMask(GL_FALSE);
            gl.glDisable(GL_DEPTH_TEST);
            gl.glDisable(GL_LIGHTING);
            // set a nice default projection
            gl.glMatrixMode(GL_PROJECTION);
            gl.glLoadIdentity();
            gl.glOrtho(-1.0, (double)width, -1.0, (double)height, -1.0, 1.0);
            // switch back to modelview matrix manipulation
            gl.glMatrixMode(GL_MODELVIEW);
            gl.glLoadIdentity();

            // draw a polygon of the correct color
            gl.glColor3f(root.bgRed, root.bgGreen, root.bgBlue);
   //         gl.glRasterPos2d(10.0, 10.0);
            root.glut.glutBitmapString(GLUTEnum.GLUT_BITMAP_HELVETICA_18, "Test") ;
            root.glut.glutStrokeString(GLUTEnum.GLUT_STROKE_ROMAN, "Howdie!");
            // reenable the depth testing and lighting
            gl.glDepthMask(GL_TRUE);
            gl.glEnable(GL_DEPTH_TEST);
            gl.glEnable(GL_LIGHTING);*/
        }
        
        // if the window has no dynamic properties, stop animation.  Otherwise, start it.
       // System.out.println("dynamic = " + dynamic + ", FPS = " + getFps() + ", isSusp = " + isSuspended());
        if (dynamic == isSuspended())
        {
            setSuspended(!dynamic);
            //System.out.println("Framerate for last animated segment: " + getFps() + " fps, max of " + getMaxFps());
        }

       // System.out.println("display end: " + glj);
        // Swap buffers
        glj.gljSwap();
        glj.gljFree();
    }      
}
