import java.util.*;
import gl4java.*;
import gl4java.utils.glut.*;
import gl4java.utils.glut.fonts.*;

/**
 * The base class for all joogl objects, which also serves to maintain
 * the state of the joogl system.
 * <p>
 * The object class keeps track of the set of jooglCanvases, and maintains a
 * handle on the gl, glu and glut classes.  This allows any joogl object to
 * request all windows be redrawn, and to issue OpenGL commands.
 * <p>
 * The object class also maintains a "sharedGLContext" variable.  This will 
 * be the GLContext of the first window to be created, and is used to set the
 * sharedGLContext of all subsequent windows.  By sharing the GLcontext, the
 * windows can shared display lists (and textures, if we ever add textures
 * to joogl).
 *
 * @version 	1.00
 * @author      Blair MacIntyre
 */
public class jooglObject implements GLEnum, GLUEnum, Cloneable {
    /** The set of all canvases */
    static private Vector canvases = null;
    /** Has the joogl system been initialized? */
    static private boolean init = false;

    /** Used for the interface to GL4Java.  Allows all joogle objects to make gl calls. */
    static public GLFunc gl = null;
    /** Used for the interface to GL4Java.  Allows all joogle objects to make glu calls. */
    static public GLUFunc glu = null;
    /** Used for the interface to GL4Java.  Allows all joogle objects to make glut calls. */
    static public GLUTFunc glut = null;
    
    /** Save the context of the first registered canvas, so it can be specified
     * as a shared context for the rest of the canvases. */
    static protected GLContext sharedGLContext = null;
        
    /** The joogl object constructor, which initializes the joogl system if not yet done. */
    public jooglObject() {
        if (!init) {
            init = true;
            canvases = new Vector();
        }
    }

    /** Add a canvas to joogl.
     * <p>
     * Each canvas calls this in its init() method to add itself to the set of known joogl canvases.
     *
     * @param canvas the jooglCanvas
     */
    public void addCanvas (jooglCanvas canvas)
    {
        canvases.add(canvas);
        if (gl == null)
        {
            gl = canvas.getGL();
            glu = canvas.getGLU();
            glut = new GLUTFuncLightImplWithFonts(gl, glu);
        }
    }
    
    /** Remove a canvas from joogl.
     *
     * Each canvas calls this in its cvsDispose() method to remove itself to the set of known joogl canvases.
     *
     * @param canvas the jooglCanvas
     */
    public void removeCanvas (jooglCanvas canvas)
    {
        canvases.remove(canvas);
    }
    
    /** Set the shared GLContext.
     *
     * If a canvas sees that the shared context is not set, it sets it with its own context.
     *
     * @param glj a GLContext to use as the shared context
     */
    public void setSharedContext(GLContext glj)
    {
        sharedGLContext = glj;
    } 
    
    /** Get the shared GLContext.
     *
     * @return the shared GLContext.
     */
    public GLContext getSharedContext()
    {
        return sharedGLContext;
    } 
    
    /*************************************************
    * internal functions
    *************************************************/
    /** Redraw all the known jooglCanvases.
     * 
     * Each time we change a node, redraw everything.  Ideally, we would only 
     * redraw the roots that contained this node, but we do not do that (for simplicity).
     */
    protected void redrawWindows ()
    {
        //System.out.println("redrawWindows");
        Iterator i = canvases.iterator();
        while (i.hasNext()) 
            ((jooglCanvas)(i.next())).repaint();
    }

}
