/**
 * The geometry node for a polygon.
 *<p>
 * A polygon is specified by a list of 3D points specified in the vertex 
 * parameter.  There are numVertex*3 elements in the array, where the
 * elements are [x1,y1,z1,x2,y2,z2,....].
 * The polygon is automatically closed, so vertex[numVertex] is connected to 
 * vertex[0] to close the polygon (ie. a unit square in the x-y plane
 * would be created with:
 * <pre>
 * float[] vertex = {0.0,0.0,0.0, 1.0,0.0,0.0, 1.0,1.0,0.0, 0.0,1.0,0.0};
 * jooglPolygon poly = new jooglPolygon(vertex);
 *</pre>
 * ALERT: make sure you create a "two-sided polygon", since we have
 * backface culling enabled.  You do this by drawing two polygons, one
 * using the points in the order given, one using them in reverse order.
 *
 * @version 	1.00
 * @author      Blair MacIntyre
 */
public class jooglPolygon extends jooglGeometry implements Cloneable{
    /** The number of vertices in the polygon. */
    protected int numVertex;
    /** The vertices. */
    protected float[] vertex;

    /** A constructor.
     *
     * @param v the vector of 3*n floating point values for an n-vertex polygon.
     */
    public jooglPolygon(float[] v)
    {
        super();
        numVertex = v.length/3;
        vertex = new float[v.length];
        for (int i = 0; i<v.length; i++)  
            vertex[i] = v[i];
    }
    
    /**
     * Creates a new jooglPolygon object with fields
     * identical to this object's.
     */
    public jooglNode snapshot(long time) {
    	jooglPolygon ret;
    	try{
    		ret = (jooglPolygon)clone();
    	} catch(CloneNotSupportedException e) { return null; }
    		
    	ret.vertex = (float [])vertex.clone();
    	
	if(material != null)
	        ret.setMaterial(material.snapshot(time));
	       
	if(transform != null)
	    	ret.setTransform(transform.snapshot(time));

	if(drawMode != null)
	    	ret.setDrawMode(drawMode.snapshot(time));
	    	
	if(shadeMode != null)
	    	ret.setShadeMode(shadeMode.snapshot(time));

	if(pickable != null)
	    	ret.setPickable(pickable.snapshot(time));

    	return ret;
    }

    /** Internal function, to recreate the polygon display list */
    protected void recreateDisplayList()
    {	
        numVertex = vertex.length/3;

	gl.glNewList(dl, GL_COMPILE);
	gl.glBegin(GL_POLYGON);
        for (int i = 0; i < numVertex; i++)
	  gl.glVertex3f(vertex[i*3], vertex[i*3+1], vertex[i*3+2]);
	gl.glEnd();
	gl.glBegin(GL_POLYGON);
        for (int i = numVertex-1; i>=0; i--)
	  gl.glVertex3f(vertex[i*3], vertex[i*3+1], vertex[i*3+2]);
	gl.glEnd();
	gl.glEndList();
    }
    
}
