/**
 * The spot light node.
 *
 * Creates an OpenGL spot light.  Read the OpenGL documentation carefully on the light parameters.
 *
 * @version 	1.00
 * @author      Blair MacIntyre
 */
public class jooglSpotLight extends jooglLight {
    /** the light position */
    protected float[] pos = new float[4];
    /** the light direction */
    protected float[] dir = new float[3];
    /** The constant attenuation factor. */
    protected float ca;
    /** The linear attenuation factor. */
    protected float la;
    /** The quadratic attenuation factor. */
    protected float qa;

    /** the cutoff for the spot spread */
    protected float cutoff;
    /** the falloff exponent for the spot spread */
    protected float exponent;
    
    /**
     * Create a spot light with a certain (r,g,b) color, position, direction, attenuation, cutoff and exponent.
     *
     * @param r		the red component of the light
     * @param g		the green component of the light
     * @param b		the blue component of the light
     * @param px	the x component of the position value for the light
     * @param py	the y component of the position value for the light
     * @param pz	the z component of the position value for the light
     * @param dx	the x component of the direction value for the light
     * @param dy	the y component of the direction value for the light
     * @param dz	the z component of the direction value for the light
     * @param _la	The constant attenuation factor
     * @param _ca	The linear attenuation factor
     * @param _qa	The quadratic attenuation factor
     * @param c		The cutoff for the spread
     * @param e		The exponent for the falloff for the spread
     */
    public jooglSpotLight(float r, float g, float b,
                          float px, float py, float pz,
                          float dx, float dy, float dz,
                          float _ca, float _la, float _qa,
                          float c, float e)
    {
        super(r, g, b);
        pos[0] = px;
        pos[1] = py;
        pos[2] = pz;
        pos[3] = 1.0f;
        dir[0] = dx;
        dir[1] = dy;
        dir[2] = dz;
        ca = _ca;
        la = _la;
        qa = _qa;
        cutoff = c;
        exponent = e;
    } 
       
    /**
     * Create a spot light with a certain (r,g,b) color, position, direction, attenuation, cutoff and exponent.
     *
     * @param r		the red component of the light
     * @param g		the green component of the light
     * @param b		the blue component of the light
     * @param px	the x component of the position value for the light
     * @param py	the y component of the position value for the light
     * @param pz	the z component of the position value for the light
     * @param dx	the x component of the direction value for the light
     * @param dy	the y component of the direction value for the light
     * @param dz	the z component of the direction value for the light
     * @param _la	The constant attenuation factor
     * @param _ca	The linear attenuation factor
     * @param _qa	The quadratic attenuation factor
     * @param c		The cutoff for the spread
     * @param e		The exponent for the falloff for the spread
     */
    public jooglSpotLight(double r, double g, double b,
                          double px, double py, double pz,
                          double dx, double dy, double dz,
                          double _ca, double _la, double _qa,
                          double c, double e)
    {
        super((float)r, (float)g, (float)b);
        pos[0] = (float)px;
        pos[1] = (float)py;
        pos[2] = (float)pz;
        pos[3] = 1.0f;
        dir[0] = (float)dx;
        dir[1] = (float)dy;
        dir[2] = (float)dz;
        ca = (float)_ca;
        la = (float)_la;
        qa = (float)_qa;
        cutoff = (float)c;
        exponent = (float)e;
    } 

    /**
     * Create a new jooglSpotLight object with fields
     * identical to this object's.
     */
    public jooglNode snapshot(long time) {
    	jooglSpotLight ret = new jooglSpotLight(r, g, b, pos[0], pos[1], pos[2], dir[0], dir[1], dir[2], ca, la, qa, cutoff, exponent);
    	
	if(material != null)
	        ret.setMaterial(material.snapshot(time));
	       
	if(transform != null)
	    	ret.setTransform(transform.snapshot(time));

	if(drawMode != null)
	    	ret.setDrawMode(drawMode.snapshot(time));
	    	
	if(shadeMode != null)
	    	ret.setShadeMode(shadeMode.snapshot(time));

	if(pickable != null)
	    	ret.setPickable(pickable.snapshot(time));
    	
	return ret;
    }
           
    /**
     * Set the light direction.
     *
     * @param dx	the x component of the direction value for the light
     * @param dy	the y component of the direction value for the light
     * @param dz	the z component of the direction value for the light
     */
    public void setDirection(float dx, float dy, float dz)
    {
        dir[0] = dx;
        dir[1] = dy;
        dir[2] = dz;
        if (lightOn)
            redrawWindows();
    }    
    
    /**
     * Retrive the current light direction.
     *
     * @return a 3 element array containing the direction of the light.
     */
    public float[] getDirection()
    {
        float [] d = new float[3];
        d[0] = dir[0];
        d[1] = dir[1];
        d[2] = dir[2];
        return d;
    }    
    
    /**
     * Set the light position.
     *
     * @param px	the x component of the position value for the light
     * @param py	the y component of the position value for the light
     * @param pz	the z component of the position value for the light
     */
    public void setPosition(float px, float py, float pz)
    {
        pos[0] = px;
        pos[1] = py;
        pos[2] = pz;
        if (lightOn)
            redrawWindows();
    }    
    
    /**
     * Retrive the current light position.
     *
     * @return a 3 element array containing the position of the light.
     */
    public float[] getPosition()
    {
        float[] p = new float[3];
        p[0] = pos[0];
        p[1] = pos[1];
        p[2] = pos[2];
        return p;
    }    

    /**
     * Set the light attenuation.
     *
     * @param la	The constant attenuation factor
     * @param ca	The linear attenuation factor
     * @param qa	The quadratic attenuation factor
     */
    public void setAttenuation(float _ca, float _la, float _qa)
    {
        ca = _ca;
        la = _la;
        qa = _qa;
        if (lightOn)
            redrawWindows();
    }    
    
    /**
     * Retrive the current light attenuation.
     *
     * @return a 3 element array containing the attenuation of the light.
     */
    public float[] getAttenuation()
    {
        float[] a = new float[3];
        a[0] = ca;
        a[1] = la;
        a[2] = qa;
        return a;
    }    

    /** Set the cutoff value.
     * 
     * @param c the new cutoff value
     */
    public void setCutoff(float c)
    {
        cutoff = c;
        if (lightOn)
            redrawWindows();
    }
    
    /** Get the current cutoff value.
     * 
     * @return the current cutoff value
     */
    public float getCutoff()
    {
        return cutoff;
    }
    
    /** Set the exponent value.
     * 
     * @param e the new exponent value
     */
    public void setExponent(float e)
    {
        exponent = e;
        if (lightOn)
            redrawWindows();
    }
    
    /** Get the current exponent value.
     * 
     * @return the current exponent value
     */
    public float getExponent()
    {
        return exponent;
    }
    
    /**
     * Internal Method:  called to create the OpenGL spot light.
     *
     * This method will be called each time a light is encountered in the scene graph to turn an
     * OpenGL spot light on based on the contents of the node.   This method will only turn on
     * the light if its "lightOn" boolean is true, and will return the value of lightOn.
     *
     * @param lightNum	the OpenGL light number to use for this light.
     *
     @ @return a boolean indicating if the light is on or off.
     */
    protected boolean activate(int lightNum)
    {
        if (lightOn) {
	    gl.glLightfv(lightNum, GL_AMBIENT, zeroLight);
	    gl.glLightfv(lightNum, GL_DIFFUSE, lm);
	    gl.glLightfv(lightNum, GL_SPECULAR, zeroLight);
	    gl.glLightfv(lightNum, GL_POSITION, pos);
	    gl.glLightf(lightNum, GL_CONSTANT_ATTENUATION, ca);
	    gl.glLightf(lightNum, GL_LINEAR_ATTENUATION, la);
	    gl.glLightf(lightNum, GL_QUADRATIC_ATTENUATION, qa);
	    gl.glLightf(lightNum, GL_SPOT_CUTOFF, cutoff);
	    gl.glLightf(lightNum, GL_SPOT_EXPONENT, exponent);
	    gl.glLightfv(lightNum, GL_SPOT_DIRECTION, dir);

            gl.glEnable(lightNum);
            return true;
        }
        return false;
    }    
}
